use std::ops::Div;
use std::path::{Path, PathBuf};
use std::process::exit;

use anyhow::Result;

use ark_bls12_381::Bls12_381;
use ark_ec::pairing::Pairing;
use ark_ff::PrimeField;
use ark_poly::univariate::DensePolynomial;
use ark_poly::DenseUVPolynomial;
use ark_poly_commit::kzg10::Powers;
use ark_serialize::{CanonicalDeserialize, Compress, Validate};
use tracing::{info, warn};

use komodo::{
    encode,
    fec::{decode, Shard},
    fs,
    linalg::Matrix,
    recode, setup, verify, Block,
};

type UniPoly12_381 = DensePolynomial<<Bls12_381 as Pairing>::ScalarField>;

const COMPRESS: Compress = Compress::Yes;
const VALIDATE: Validate = Validate::Yes;

#[allow(clippy::type_complexity)]
fn parse_args() -> (
    Vec<u8>,
    usize,
    usize,
    bool,
    String,
    bool,
    bool,
    bool,
    bool,
    usize,
    String,
    Vec<String>,
) {
    let bytes_path = std::env::args()
        .nth(1)
        .expect("expected path to bytes as first positional argument");
    let bytes = if bytes_path.is_empty() {
        vec![]
    } else {
        std::fs::read(bytes_path).unwrap()
    };
    let k: usize = std::env::args()
        .nth(2)
        .expect("expected k as second positional argument")
        .parse()
        .expect("could not parse k as an int");
    let n: usize = std::env::args()
        .nth(3)
        .expect("expected n as third positional argument")
        .parse()
        .expect("could not parse n as an int");
    let do_generate_powers: bool = std::env::args()
        .nth(4)
        .expect("expected do_generate_powers as fourth positional argument")
        .parse()
        .expect("could not parse do_generate_powers as a bool");
    let home_dir = std::env::args()
        .nth(5)
        .expect("expected home_dir as fifth positional argument");
    let do_reconstruct_data: bool = std::env::args()
        .nth(6)
        .expect("expected do_reconstruct_data as sixth positional argument")
        .parse()
        .expect("could not parse do_reconstruct_data as a bool");
    let do_verify_blocks: bool = std::env::args()
        .nth(7)
        .expect("expected do_verify_blocks as seventh positional argument")
        .parse()
        .expect("could not parse do_verify_blocks as a bool");
    let do_combine_blocks: bool = std::env::args()
        .nth(8)
        .expect("expected do_combine_blocks as eigth positional argument")
        .parse()
        .expect("could not parse do_combine_blocks as a bool");
    let do_inspect_blocks: bool = std::env::args()
        .nth(9)
        .expect("expected do_inspect_blocks as ninth positional argument")
        .parse()
        .expect("could not parse do_inspect_blocks as a bool");
    let nb_bytes: usize = std::env::args()
        .nth(10)
        .expect("expected nb_bytes as 10th positional argument")
        .parse()
        .expect("could not parse nb_bytes as a usize");
    let encoding_method = std::env::args()
        .nth(11)
        .expect("expected encoding_method as 11th positional argument");
    let block_hashes = std::env::args().skip(12).collect::<Vec<_>>();

    (
        bytes,
        k,
        n,
        do_generate_powers,
        home_dir,
        do_reconstruct_data,
        do_verify_blocks,
        do_combine_blocks,
        do_inspect_blocks,
        nb_bytes,
        encoding_method,
        block_hashes,
    )
}

fn throw_error(code: i32, message: &str) {
    eprint!("{}", message);
    exit(code);
}

pub fn generate_random_powers<E, P>(
    n: usize,
    powers_dir: &Path,
    powers_filename: Option<&str>,
) -> Result<()>
where
    E: Pairing,
    P: DenseUVPolynomial<E::ScalarField, Point = E::ScalarField>,
    for<'a, 'b> &'a P: Div<&'b P, Output = P>,
{
    info!("generating new powers");
    let powers = setup::random::<E, P>(n)?;

    fs::dump(&powers, powers_dir, powers_filename, COMPRESS)?;

    Ok(())
}

pub fn verify_blocks<E, P>(
    blocks: &[(String, Block<E>)],
    powers: Powers<E>,
) -> Result<(), ark_poly_commit::Error>
where
    E: Pairing,
    P: DenseUVPolynomial<E::ScalarField, Point = E::ScalarField>,
    for<'a, 'b> &'a P: Div<&'b P, Output = P>,
{
    let res = blocks
        .iter()
        .map(|(f, b)| Ok((f, verify::<E, P>(b, &powers)?)))
        .collect::<Result<Vec<(&String, bool)>, ark_poly_commit::Error>>()?;

    eprint!("[");
    for (f, v) in res {
        eprint!("{{block: {:?}, status: {}}}", f, v);
    }
    eprint!("]");
    Ok(())
}

fn main() {
    tracing_subscriber::fmt::try_init().expect("cannot init logger");

    let (
        bytes,
        k,
        n,
        do_generate_powers,
        home_dir,
        do_reconstruct_data,
        do_verify_blocks,
        do_combine_blocks,
        do_inspect_blocks,
        nb_bytes,
        encoding_method,
        block_hashes,
    ) = parse_args();

    let home_dir = PathBuf::from(&home_dir);
    let block_dir = home_dir.join("blocks/");
    let powers_dir = home_dir;
    let powers_filename = "powers";
    let powers_file = powers_dir.join(powers_filename);

    if do_generate_powers {
        generate_random_powers::<Bls12_381, UniPoly12_381>(
            nb_bytes,
            &powers_dir,
            Some(powers_filename),
        )
        .unwrap_or_else(|e| throw_error(1, &format!("could not generate powers: {}", e)));

        exit(0);
    }

    if do_reconstruct_data {
        let blocks: Vec<Shard<Bls12_381>> =
            fs::read_blocks::<Bls12_381>(&block_hashes, &block_dir, COMPRESS, VALIDATE)
                .unwrap_or_else(|e| {
                    throw_error(1, &format!("could not read blocks: {}", e));
                    unreachable!()
                })
                .iter()
                .cloned()
                .map(|b| b.1.shard)
                .collect();
        eprintln!(
            "{:?}",
            decode::<Bls12_381>(blocks).unwrap_or_else(|e| {
                throw_error(1, &format!("could not decode: {}", e));
                unreachable!()
            })
        );

        exit(0);
    }

    if do_combine_blocks {
        let blocks = fs::read_blocks::<Bls12_381>(&block_hashes, &block_dir, COMPRESS, VALIDATE)
            .unwrap_or_else(|e| {
                throw_error(1, &format!("could not read blocks: {}", e));
                unreachable!()
            });

        let formatted_output = fs::dump_blocks(
            &[
                recode(&blocks.iter().map(|(_, b)| b).cloned().collect::<Vec<_>>())
                    .unwrap_or_else(|e| {
                        throw_error(1, &format!("could not encode block: {}", e));
                        unreachable!()
                    })
                    .unwrap_or_else(|| {
                        throw_error(1, "could not recode block (list of blocks is likely empty)");
                        unreachable!()
                    }),
            ],
            &block_dir,
            COMPRESS,
        )
        .unwrap_or_else(|e| {
            throw_error(1, &format!("could not dump block: {}", e));
            unreachable!()
        });

        eprint!("{}", formatted_output);

        exit(0);
    }

    if do_inspect_blocks {
        let blocks = fs::read_blocks::<Bls12_381>(&block_hashes, &block_dir, COMPRESS, VALIDATE)
            .unwrap_or_else(|e| {
                throw_error(1, &format!("could not read blocks: {}", e));
                unreachable!()
            });
        eprint!("[");
        for (_, block) in &blocks {
            eprint!("{},", block);
        }
        eprintln!("]");

        exit(0);
    }

    info!("reading powers from file `{:?}`", powers_file);
    let powers = if let Ok(serialized) = std::fs::read(&powers_file) {
        info!("deserializing the powers from `{:?}`", powers_file);
        Powers::<Bls12_381>::deserialize_with_mode(&serialized[..], COMPRESS, VALIDATE)
            .unwrap_or_else(|e| {
                throw_error(
                    1,
                    &format!("could not deserialize powers from {:?}: {}", powers_file, e),
                );
                unreachable!()
            })
    } else {
        warn!("could not read powers from `{:?}`", powers_file);
        info!("regenerating temporary powers");
        setup::random::<Bls12_381, UniPoly12_381>(nb_bytes).unwrap_or_else(|e| {
            throw_error(1, &format!("could not generate powers: {}", e));
            unreachable!()
        })
    };

    if do_verify_blocks {
        verify_blocks::<Bls12_381, UniPoly12_381>(
            &fs::read_blocks::<Bls12_381>(&block_hashes, &block_dir, COMPRESS, VALIDATE)
                .unwrap_or_else(|e| {
                    throw_error(1, &format!("could not read blocks: {}", e));
                    unreachable!()
                }),
            powers,
        )
        .unwrap_or_else(|e| {
            throw_error(1, &format!("Failed to verify blocks: {}", e));
            unreachable!()
        });

        exit(0);
    }

    let encoding_mat = match encoding_method.as_str() {
        "vandermonde" => {
            let points: Vec<<Bls12_381 as Pairing>::ScalarField> = (0..n)
                .map(|i| {
                    <Bls12_381 as Pairing>::ScalarField::from_le_bytes_mod_order(&i.to_le_bytes())
                })
                .collect();
            Matrix::vandermonde(&points, k)
        }
        "random" => Matrix::random(k, n),
        m => {
            throw_error(1, &format!("invalid encoding method: {}", m));
            unreachable!()
        }
    };

    let formatted_output = fs::dump_blocks(
        &encode::<Bls12_381, UniPoly12_381>(&bytes, &encoding_mat, &powers).unwrap_or_else(|e| {
            throw_error(1, &format!("could not encode: {}", e));
            unreachable!()
        }),
        &block_dir,
        COMPRESS,
    )
    .unwrap_or_else(|e| {
        throw_error(1, &format!("could not dump blocks: {}", e));
        unreachable!()
    });

    eprint!("{}", formatted_output);
}
